package be.iminds.ilabt.jfed.ssh_terminal_tool;

import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.PuTTYFilesKeyInfo;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshFilesKeyInfo;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfoFactory;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.preferences.ProxyPreferencesManager;
import be.iminds.ilabt.jfed.ssh_terminal_tool.putty.PuttyHelper;
import be.iminds.ilabt.jfed.ssh_terminal_tool.unix.OpenSshKnownHostsFile;
import be.iminds.ilabt.jfed.ssh_terminal_tool.unix.SshAgentHelper;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.util.LocalForwardUtil;
import be.iminds.ilabt.jfed.util.OSDetector;
import be.iminds.ilabt.jfed.util.tmp_file_helpers.PermanentFile;
import be.iminds.ilabt.jfed.util.tmp_file_helpers.TmpContentFile;
import be.iminds.ilabt.jfed.util.tmp_file_helpers.TmpFile;
import com.google.common.net.InetAddresses;
import javafx.scene.control.*;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Singleton;
import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.regex.Pattern;

@Singleton
public class ExternalRdpApp {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(ExternalRdpApp.class);

    private final GeniUserProvider geniUserProvider;
    private final JFedGuiPreferences jFedPreferences;
    private final ProxyPreferencesManager proxyPreferencesManager;
    private final LocalForwardUtil localForwardUtil;

    @Inject
    public ExternalRdpApp(GeniUserProvider geniUserProvider,
                          ProxyPreferencesManager proxyPreferencesManager,
                          JFedGuiPreferences jFedPreferences,
                          LocalForwardUtil localForwardUtil) {
        this.geniUserProvider = geniUserProvider;
        this.proxyPreferencesManager = proxyPreferencesManager;
        this.jFedPreferences = jFedPreferences;
        this.localForwardUtil = localForwardUtil;
    }

    public boolean isConfigured() {
        //TODO rewrite
        LOG.debug("ExternalRdpApp.isConfigured");
        if (OSDetector.os == OSDetector.OS.WIN) {
            File puttyDir = jFedPreferences.getFile(GuiPreferenceKey.PREF_PUTTY_DIRECTORY);
            LOG.debug("ExternalRdpApp.isConfigured WIN puttyDir=" + puttyDir);
            if (puttyDir == null) return false;
            boolean valid = JFedGuiPreferences.isValidPuttyDir(puttyDir);
            LOG.debug("ExternalRdpApp.isConfigured WIN puttyDir=" + puttyDir + " valid=" + valid);
            return valid;
        }

        if (OSDetector.os == OSDetector.OS.UNIX) {
            return jFedPreferences.getString(GuiPreferenceKey.PREF_COMMAND_UNIX_TERMINAL) != null;
        }

        if (OSDetector.os == OSDetector.OS.MAC) {
            return jFedPreferences.getString(GuiPreferenceKey.PREF_COMMAND_UNIX_TERMINAL) != null;
        }

        return true;
    }

    public void launch(@Nullable String targetServerUserName, @Nonnull String hostName, int port,
                       @Nullable String password,
                       @Nullable JFedConnection.ProxyInfo customProxyInfo) {
        final GeniUser geniUser = geniUserProvider.getLoggedInGeniUser();

        //This will set the proxyInfo to what is needed, based on if proxy exists, preferences and connectivity data.
        JFedConnection.SshProxyInfo jFedProxyInfo = geniUser.getUserAuthorityServer() == null ? null :
                proxyPreferencesManager.getSshProxySettings(geniUser.getUserAuthorityServer().getTestbed().getProxies(), geniUser);

        JFedConnection.SshProxyInfo usedProxyInfo = jFedProxyInfo;
        if (customProxyInfo != null) {
            usedProxyInfo = (JFedConnection.SshProxyInfo) customProxyInfo;
        }

        if (usedProxyInfo == null)
            LOG.info("ExternalRdpApp: No proxy will be used.");
        else {
            LOG.info("ExternalRdpApp: Proxy will be used: " + usedProxyInfo);

            //setup connection to server over the proxy, forwarding to a local port
            int forwardPort = 0;
            try {
                forwardPort = localForwardUtil.forwardOverSshProxy(usedProxyInfo, hostName, port);
            } catch (IOException e) {
                throw new RuntimeException("Could not setup SSH port forwarding", e);
            }

            //Now continue, but make the connection to localhost instead
            hostName = "localhost";
            port = forwardPort;
        }

        if (OSDetector.os == OSDetector.OS.WIN) {
            launchWindowsRemoteDesktop(
                    targetServerUserName,
                    hostName,
                    port,
                    password);
            return;
        }

//        if (OSDetector.os == OSDetector.OS.UNIX || OSDetector.os == OSDetector.OS.MAC) {
        launchShellRdpApp(
                targetServerUserName,
                hostName,
                port,
                password);
        return;
//        }
//
//        throw new RuntimeException("No Rdp App known for OS \"" + OSDetector.os + "\"");
    }

    private void showWindowsRemoteDesktopDetails(@Nullable String targetServerUserName,
                                                 @Nonnull String hostName,
                                                 int port,
                                                 @Nullable String password) {
        VBox pane = new VBox();
        pane.setPrefWidth(750);
        pane.setSpacing(5);

        String text = String.format("You can open a windows Remote Desktop connection to %s:%d\n",
                hostName, port);
        if (targetServerUserName != null || password != null)
            text += "You can login ";
        if (targetServerUserName != null)
            text += String.format(" with username \"%s\"", targetServerUserName);
        if (password != null)
            text += String.format(" using the password \"%s\"", password);

        Label textLabel = new Label(text);
        textLabel.setPrefHeight(75);

        textLabel.setWrapText(true);
        pane.getChildren().add(textLabel);

        DialogPane dialogPane = new DialogPane();
        dialogPane.setContent(pane);
        dialogPane.getButtonTypes().setAll(ButtonType.OK);


        Dialog<ButtonType> dialog = new Dialog<>();
        dialog.setTitle("Remote Desktop Info");
        dialog.setDialogPane(dialogPane);

        dialog.showAndWait();
    }

    private void launchWindowsRemoteDesktop(@Nullable String targetServerUserName,
                                            @Nonnull String hostName,
                                            int port,
                                            @Nullable String password) {
        //TODO: implement starting remote desktop on windows
        showWindowsRemoteDesktopDetails(targetServerUserName, hostName, port, password);
    }

    private void launchShellRdpApp(@Nullable String targetServerUserName,
                                   @Nonnull String hostName,
                                   int port,
                                   @Nullable String password) {
        //TODO: implement starting remote desktop on unix based OSes (using custom command in jFed preferences)
        //rdesktop -p <pass> -u <username> <server>:<port>
        showWindowsRemoteDesktopDetails(targetServerUserName, hostName, port, password);
    }
}

